"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var express = require("express");
var http = require("http");
var https = require("https");
var path = require("path");
var fs = require("fs");
var tungus = require('tungus');
var mongoose = require("mongoose");
var ip = require("ip");
var util = require("./util");
var multer = require('multer');
// express middleware
var request = require('request');
var progress = require('progress-stream');
var expressLogger = require('morgan');
var bodyParser = require('body-parser');
var methodOverride = require('method-override');
var engine = require('ejs-mate');
var cookieParser = require('cookie-parser');
var cookieSession = require('cookie-session');
var app = express();
global["app"] = app;
// import models
var Config = require("./models/Config");
var Device = require("./models/Device");
var DeviceInfo = require("./models/DeviceInfo");
var Log = require("./models/Log");
var Schedule = require("./models/Schedule");
var DeviceGroup = require("./models/DeviceGroup");
var devices = {};
var repositoryFirmwares = [];
var configs = [];
var version = fs.readFileSync("version.txt", "utf8").replace('\r', '').replace('\n', '');
if ('development' == app.get('env')) {
    global["datadir"] = __dirname;
}
else {
    if (process.platform == 'darwin') {
        global["datadir"] = process.env.HOME + 'Library/Preferences/Netonix Manager';
    }
    else if (process.platform == 'win32') {
        global["datadir"] = process.env.APPDATA + "/Netonix Manager";
    }
    else {
        // linux
        global["datadir"] = '/var/lib/netonix-manager';
    }
    if (fs.existsSync(global["datadir"]) == false) {
        fs.mkdirSync(global["datadir"]);
    }
    if (fs.existsSync(global["datadir"] + "/firmwares") == false) {
        fs.mkdirSync(global["datadir"] + "/firmwares");
    }
    if (fs.existsSync(global["datadir"] + "/configs") == false) {
        fs.mkdirSync(global["datadir"] + "/configs");
    }
    if (fs.existsSync(global["datadir"] + "/storage") == false) {
        fs.mkdirSync(global["datadir"] + "/storage");
    }
}
var firmwareStorage = multer.diskStorage({
    destination: function (req, file, cb) {
        cb(null, path.join(global["datadir"], 'firmwares'));
    },
    filename: function (req, file, cb) {
        cb(null, file.originalname);
    }
});
// start up DB
var db = mongoose.connection;
db.on('error', console.error);
mongoose.connect('tingodb://' + global["datadir"] + '/storage', function (err) {
    util.log("Version " + version + " starting up");
    // create default configuration if there is no configuration
    Config.findOne({}, function (err, config) {
        if (err)
            return console.log(err);
        if (config == null) {
            config = new Config();
            config.username = "admin";
            config.password = "admin";
            config.last_seen_version = version;
            config.save();
        }
    });
});
var sessionToken = "";
require('crypto').randomBytes(48, function (err, buffer) {
    sessionToken = buffer.toString('hex');
});
// all environments
app.engine('ejs', engine);
app.set('port', process.env.PORT || 3000);
app.set('views', path.join(__dirname, 'views'));
app.set('view engine', 'ejs');
app.use(expressLogger('dev'));
app.use(require('express-domain-middleware'));
app.use(cookieParser());
app.use(cookieSession({ name: 'session', secret: 'thisisnoterysecret' }));
app.use(bodyParser.json());
app.use(bodyParser.urlencoded({ extended: true }));
app.use(methodOverride());
app.use(express.static(path.join(__dirname, 'public')));
app.use(function (req, res, next) {
    // we use this for testing
    if ('development' == app.get('env') && req.header("Authorization") == "lasdhsvilnslinvgsdgf") {
        next();
        return;
    }
    if (req.url !== '/' && (!req.session || req.session.token != sessionToken)) {
        res.sendStatus(401);
        return;
    }
    next();
});
// routes
app.get('/', function (req, res) {
    var tarpit_delay = 0;
    if (req.session.tarpit_time) {
        req.session.tarpit_time = new Date(req.session.tarpit_time);
        if (new Date() >= req.session.tarpit_time) {
            delete req.session.tarpit_time;
            delete req.session.loginFailureCount;
        }
        else {
            tarpit_delay = Math.ceil((req.session.tarpit_time.getTime() - new Date().getTime()) / 1000);
        }
    }
    res.render('index', { title: 'Netonix Manager', _layoutFile: 'layout', error: '', tarpit_delay: tarpit_delay });
});
app.post('/', function (req, res, next) {
    Config.findOne({}, function (err, config) {
        if (err)
            return console.log(err);
        if (req.body.username && req.body.username === config.username && req.body.password && config.comparePassword(req.body.password)) {
            req.session.token = sessionToken;
            if (repositoryFirmwares.length == 0) {
                refreshFirmwareList();
            }
            res.redirect('/main.html');
        }
        else {
            var tarpit_delay = 0;
            util.log("Invalid login attempt for user '" + req.body.username + "'");
            if (req.session.loginFailureCount) {
                req.session.loginFailureCount++;
                if (req.session.loginFailureCount >= 5) {
                    req.session.tarpit_time = new Date(new Date().getTime() + 60000);
                    tarpit_delay = 60;
                }
            }
            else {
                req.session.loginFailureCount = 1;
            }
            res.render('index', { title: 'Netonix Manager', _layoutFile: 'layout', error: 'Invalid username or password', tarpit_delay: tarpit_delay });
        }
    });
});
app.get('/logout', function (req, res, next) {
    delete req.session.token;
    res.redirect('/');
});
app.get('/api/version', function (req, res) {
    res.json({ version: version, crash_log: fs.existsSync("crash.log") });
});
app.post('/api/crashlog/:op', function (req, res) {
    var logdata = fs.readFileSync("crash.log");
    fs.unlinkSync("crash.log");
    request.post('https://forum.netonix.com/netonix_manager_crash_report.php', { form: { message: logdata, submit: 'submit', version: version } }, function (error, response, body) {
    });
    res.json({});
});
function getDevice(device) {
    var newObject = JSON.parse(JSON.stringify(device));
    delete newObject.deviceInfo;
    return newObject;
}
app.get('/api/devices', function (req, res) {
    var result = Array();
    Config.findOne({}, function (err, config) {
        if (!err) {
            for (var key in devices) {
                if (config.selected_group.trim() == "all") {
                    result.push(getDevice(devices[key]));
                }
                else if (devices[key].device_group.trim() == config.selected_group.trim()) {
                    result.push(getDevice(devices[key]));
                }
            }
            var total = 0;
            if (scanStart > 0) {
                total = scanEnd - scanStart + 1;
            }
            res.json({ devices: result, scanProgress: scanProgress, scanTotal: total, scanFound: scanFound });
        }
        else {
            util.log("Error getting config in /api/devices");
            var total = 0;
            if (scanStart > 0) {
                total = scanEnd - scanStart + 1;
            }
            res.json({ devices: result, scanProgress: scanProgress, scanTotal: total, scanFound: scanFound });
        }
    });
});
app.post('/api/devices', function (req, res) {
    DeviceInfo.findOne({ address: req.body.device.address, ssh_port: req.body.device.ssh_port }, function (err, device) {
        if (device == null) {
            DeviceInfo.create(req.body.device, function (err, device) {
                if (err) {
                    // this will happen if the record already exists
                    res.sendStatus(409);
                }
                else {
                    devices[device.id] = new Device(device);
                    res.redirect("/api/devices/" + device.id);
                    util.log("Created device " + device.address);
                }
            });
        }
        else {
            // this will happen if the device already exists
            res.sendStatus(409);
        }
    });
});
app.get('/api/devices/:id', function (req, res) {
    res.json(getDevice(devices[req.params.id]));
});
app.put('/api/devices/:id', function (req, res) {
    DeviceInfo.update({ _id: req.params.id }, req.body.device, function (err, numAffected) {
        if (err) {
            // if the id doesn't exist
            res.sendStatus(404);
        }
        else {
            DeviceInfo.findById(req.params.id, function (err, deviceinfo) {
                devices[req.params.id] = new Device(deviceinfo);
                util.log("Updated device " + devices[req.params.id].address);
                res.sendStatus(204); // success, no content
            });
        }
    });
});
app.delete('/api/devices/:id', function (req, res) {
    DeviceInfo.remove({ _id: req.params.id }, function (err) {
        if (err) {
            // if the id doesn't exist
            res.sendStatus(404);
        }
        else {
            util.log("Deleted device " + devices[req.params.id].address);
            delete devices[req.params.id];
            res.sendStatus(204); // success, no content
        }
    });
});
app.get('/api/devicelogalerts/:id', function (req, res) {
    devices[req.params.id].getLogAlerts(function (log) {
        res.json({ log: log });
    });
});
app.get("/api/getselectedgroup", function (req, res) {
    Config.findOne({}, function (err, config) {
        if (!err) {
            res.json({
                selected_group: config.selected_group
            });
        }
        else {
            util.log("Error getting the selected group from settings");
        }
    });
});
// some array handling functions
var get_item_count = function (arr, item) {
    var count = 0;
    for (var x in arr) {
        if (arr[x] == item) {
            count = count + 1;
        }
    }
    return count;
};
var distinct = function (value, index, self) {
    return self.indexOf(value) === index;
};
app.post("/api/addtogroup", function (req, res) {
    // how many devices will be added to this group?
    var new_group_count = req.body.device_ids.length;
    // in case the devices are removed from a group that is not default, add them here
    var groups_dev_removed = [];
    // make sure nothing rhetorical is happenning
    var devices_to_update = new_group_count;
    // change the devices status's to reflect
    for (var id in req.body.device_ids) {
        var dev_id = req.body.device_ids[id];
        // make sure each device doesn't already belong to the group if it does, skip it
        if (devices[dev_id].device_group == req.body.group) {
            devices_to_update--;
            continue;
        }
        // first record what devices these groups are moving from
        if (devices[dev_id].device_group != "all") {
            groups_dev_removed.push(devices[dev_id].device_group);
        }
        // now we can re-assign them
        devices[dev_id].device_group = req.body.group;
        DeviceInfo.findById(dev_id, function (err, deviceinfo) {
            deviceinfo.device_group = req.body.group.trim();
            deviceinfo.save();
        });
    }
    if (devices_to_update == 0) {
        util.log("No new group changes needed");
        res.json({ result: 'no_change' });
        return;
    }
    // now get the resulting unique items array
    var DistinctGroups = groups_dev_removed.filter(distinct);
    // next get the count difference for each
    var dev_changed = [];
    for (var i in DistinctGroups) {
        dev_changed.push(get_item_count(groups_dev_removed, DistinctGroups[i]));
    }
    // now update the group that had the devices added
    DeviceGroup.findOne({ group_name: req.body.group }, function (err, group) {
        if (!err) {
            group.group_count += new_group_count;
            group.save();
        }
        else {
            util.log("Error retrieving group info to add devices");
        }
    });
    // now update all the groups that have had devices removed
    if (dev_changed.length > 0) {
        for (var i in dev_changed) {
            DeviceGroup.findOne({ group_name: DistinctGroups[i] }, function (err, group) {
                group.group_count -= dev_changed[i];
                group.save();
            });
        }
    }
    res.json({ result: 'success' });
});
app.post("/api/setgroup/:groupname", function (req, res) {
    Config.findOne({}, function (err, config) {
        config.selected_group = req.params.groupname.trim();
        config.save();
        util.log("Changed group to: " + req.params.groupname);
        res.json({ result: 'success' });
    });
});
app.post('/api/devicegroup', function (req, res) {
    var body = req.body;
    var GroupName = body.group_name.trim();
    var GroupDescription = body.group_description;
    var NewGroup = {
        group_name: GroupName,
        group_description: GroupDescription
    };
    DeviceGroup.findOne({ group_name: GroupName }, function (err, result) {
        if (result == null) {
            DeviceGroup.create(NewGroup, function (err, result) {
                if (!err) {
                    util.log("New Group added, name: " + GroupName);
                    res.send(200);
                }
            });
        }
        else {
            util.log("group: " + GroupName + " already exists!");
            res.send(409);
        }
    });
});
app.get('/api/devicegroup', function (req, res) {
    DeviceGroup.find({}, function (err, groups) {
        if (err) {
            // shouldn't happen unless there is an issue with the database
            res.sendStatus(500);
        }
        else {
            res.json({ DeviceGroups: groups });
        }
    });
});
app.post('/api/devicegroupdelete', function (req, res) {
    var delgroup = req.body;
    for (var val in delgroup) {
        DeviceGroup.remove({ group_name: delgroup[val] }, function (err) {
            Config.findOne({}, function (c_err, config) {
                if (!c_err) {
                    config.selected_group = 'all';
                    config.save();
                    // here we need to reset all devices that may have been in this group back to default
                    DeviceInfo.find({ device_group: delgroup[val] }, function (d_err, dev_info_result) {
                        if (!d_err) {
                            for (var i = 0; i < dev_info_result.length; i++) {
                                dev_info_result[i].device_group = 'all';
                                devices[dev_info_result[i].id].device_group = 'all';
                                console.log(devices[dev_info_result[i].id].device_group);
                                dev_info_result[i].save();
                            }
                        }
                    });
                    if (!err) {
                        util.log("Deleted Device Group: " + delgroup[val].toString() + "successfully");
                        res.send(209);
                    }
                    else {
                        util.log("Failed to delete Device Group: " + delgroup[val].toString());
                        res.send(409);
                    }
                }
                else {
                    util.log("Failed to get config while removing groups");
                    res.send(409);
                }
            });
        });
    }
});
app.get('/api/firmwares', function (req, res) {
    res.json({ firmwares: fs.readdirSync(path.join(global["datadir"], 'firmwares')) });
});
app.delete('/api/firmwares/:id', function (req, res) {
    fs.unlinkSync(path.join(path.join(global["datadir"], 'firmwares'), req.params.id));
    res.sendStatus(204); // success, no content
});
app.get('/api/repositoryfirmwares', function (req, res) {
    res.json({ firmwares: repositoryFirmwares });
});
app.post('/api/uploadfirmware', multer({ storage: firmwareStorage, limits: { fileSize: 20000000, files: 1 } }).single('file'), function (req, res) {
    res.sendStatus(204).end();
});
app.get('/api/configs', function (req, res) {
    var result = [];
    var files = fs.readdirSync(path.join(global["datadir"], 'configs'));
    for (var i = 0; i < files.length; i++) {
        result.push({ name: files[i], date: fs.statSync(path.join(global["datadir"], 'configs/' + files[i])).mtime });
    }
    res.json({ configs: result });
});
app.get('/api/configs/:id', function (req, res) {
    res.sendFile(path.join(path.join(global["datadir"], 'configs'), req.params.id));
});
app.delete('/api/configs/:id', function (req, res) {
    fs.unlinkSync(path.join(path.join(global["datadir"], 'configs'), req.params.id));
    res.sendStatus(204); // success, no content
});
function upgradeDevices(body, firmwareFile) {
    if (body.scheduled) {
        var s = new Schedule();
        s.date = body.date;
        s.firmware = "Upgrade to " + firmwareFile;
        s.devices = body.devices;
        s.save();
    }
    else {
        body.devices.forEach(function (id) {
            if (devices[id] != null) {
                devices[id].upgrade_firmware_file = firmwareFile;
            }
        });
    }
}
var scanRanges = [];
var scanStart = 0;
var scanEnd = 0;
var scanLast = 0;
var scanDevice = null;
var scanFound = 0;
var scanProgress = 0;
var scanCount = 0;
function scanner() {
    if (scanCount == 0) {
        if (scanStart == 0 || scanLast > scanEnd) {
            if (scanRanges.length == 0) {
                util.log("Scan complete, found " + scanFound + " new devices");
                scanStart = 0;
                scanEnd = 0;
                scanFound = 0;
                scanLast = 0;
                scanProgress = 0;
                scanDevice = null;
                return;
            }
            var range = scanRanges.pop();
            util.log("Scanning " + range + " for devices");
            var s = range.split('/');
            var subnet;
            if (s[1].indexOf('.') > 0) {
                subnet = ip.subnet(s[0], s[1]);
            }
            else {
                subnet = ip.cidrSubnet(range);
            }
            scanStart = ip.toLong(subnet.networkAddress) + 1;
            scanEnd = ip.toLong(subnet.broadcastAddress) - 1;
            scanLast = scanStart;
            scanProgress = 0;
            scanCount = 0;
        }
        var n = 0;
        while (scanLast <= scanEnd) {
            scanCount++;
            scanAddress(ip.fromLong(scanLast), function () {
                scanProgress++;
                scanCount--;
                scanner();
            });
            scanLast++;
            n++;
            if (n == 256) {
                break;
            }
        }
    }
}
function cancelScan() {
    util.log("Scan cancelled");
    scanRanges = [];
    scanStart = 0;
    scanEnd = 0;
    scanFound = 0;
    scanLast = 0;
    scanProgress = 0;
    scanDevice = null;
}
function scanAddress(address, callback) {
    DeviceInfo.findOne({ address: address, ssh_port: scanDevice.ssh_port }, function (err, device) {
        if (device == null) {
            // very early firmwares used HTTP Basic Authentication so we need to specify the username/password
            var req = request({ url: 'https://' + scanDevice.username + ':' + scanDevice.password + '@' + address, agentOptions: { rejectUnauthorized: false }, timeout: 5000 }, function (error, response, body) {
                if (!error && response.statusCode == 200) {
                    if (body.indexOf('<i>WISP</i> Switch') > 0 || body.indexOf('Netonix') > 0) {
                        scanDevice.address = address;
                        DeviceInfo.create(scanDevice, function (err, device) {
                            if (!err) {
                                util.log("Scan found new device at " + address);
                                scanFound++;
                                devices[device.id] = new Device(device);
                            }
                        });
                    }
                }
                callback();
            });
        }
        else {
            callback();
        }
    });
}
app.post('/api/scanrange', function (req, res) {
    scanRanges = req.body.device.address.split(/[ ,]+/).map(function (item) { return item.trim(); });
    scanDevice = req.body.device;
    scanFound = 0;
    scanner();
    res.sendStatus(204).end();
});
app.delete('/api/scanrange', function (req, res) {
    cancelScan();
    res.sendStatus(204).end();
});
app.post('/api/reboot', function (req, res) {
    var body = req.body;
    if (body.scheduled) {
        var s = new Schedule();
        s.date = body.date;
        s.firmware = "Reboot";
        s.devices = body.devices;
        s.save();
    }
    else {
        body.devices.forEach(function (id) {
            if (devices[id] != null) {
                devices[id].reboot();
            }
        });
    }
    res.sendStatus(204); // success, no content
});
function setDeviceStatus(req, msg, progress) {
    req.body.devices.forEach(function (id) {
        if (devices[id] != null) {
            devices[id].status = msg;
            devices[id].statusProgress = progress;
        }
    });
}
function doDownload(req, url, errorcb) {
    var localFilename = global["datadir"] + '/firmwares/' + req.body.downloadFile;
    if (fs.existsSync(localFilename)) {
        util.log(req.body.downloadFile + " already exists, skipping download");
        upgradeDevices(req.body, req.body.downloadFile);
        return;
    }
    util.log("Downloading " + req.body.downloadFile);
    var file = fs.createWriteStream(localFilename);
    var request = http.get(url, function (response) {
        if (response.statusCode == 404) {
            file.close();
            fs.unlinkSync(localFilename);
            setDeviceStatus(req, '', -1);
            errorcb();
            return;
        }
        var str = progress({
            length: response.headers["content-length"],
            time: 100
        });
        str.on('progress', function (progress) {
            setDeviceStatus(req, "Downloading " + req.body.downloadFile, progress.percentage);
        });
        response.pipe(str).pipe(file);
        file.on('finish', function () {
            file.close();
            setDeviceStatus(req, '', -1);
            upgradeDevices(req.body, req.body.downloadFile);
        });
    }).on('error', function (err) {
        file.close();
        fs.unlinkSync(localFilename);
        setDeviceStatus(req, '', -1);
        errorcb();
    });
}
app.post('/api/upgradefirmware', function (req, res) {
    if (req.body.downloadFile) {
        doDownload(req, 'http://forum.netonix.com/firmware/' + req.body.downloadFile, function () {
            // download failed, maybe the file is in the archive directory (thanks for the typo Chris)
            doDownload(req, 'http://forum.netonix.com/firmware/archieve/' + req.body.downloadFile, function () {
                // not there either, give up
                util.log("Unable to download " + req.body.downloadFile);
            });
        });
    }
    else {
        upgradeDevices(req.body, req.body.firmwareFile);
    }
    res.sendStatus(204).end();
});
app.post('/api/testsmtp', function (req, res) {
    util.sendEmail('test email from Netonix Manager', 'This is a test email', req.body, function (errorMessage) {
        res.json({ error: errorMessage });
    });
});
app.get('/api/settings', function (req, res) {
    Config.findOne({}, function (err, config) {
        if (err)
            return console.log(err);
        var settings = JSON.parse(JSON.stringify(config));
        // for security reasons we don't want to send this
        delete settings._id;
        delete settings.password;
        res.json(settings);
    });
});
app.put('/api/settings', function (req, res) {
    Config.findOne({}, function (err, config) {
        // we can't use Config.update() here because we need to pre-save method to be called
        for (var key in config) {
            var s = req.body[key];
            if (typeof (s) !== 'undefined') {
                config[key] = req.body[key];
                // this is for testing
                if (key == 'last_seen_version') {
                    version = req.body[key];
                }
            }
        }
        config.save();
        util.log("Updated settings");
        res.sendStatus(204); // success, no content
    });
});
app.get('/api/log', function (req, res) {
    Log.find({}, function (err, results) {
        res.json({ log: results });
    });
});
app.post('/api/log', function (req, res) {
    util.log(req.body.message);
    res.redirect("/api/log");
});
app.delete('/api/log', function (req, res) {
    Log.remove({}, function (err) {
        res.sendStatus(204); // success, no content
    });
});
app.get('/api/schedules', function (req, res) {
    var results = [];
    Schedule.find({}, function (err, schedules) {
        Config.findOne({}, function (err, config) {
            if (!err) {
                for (var ri in schedules) {
                    var dev_num = schedules[ri].devices[0];
                    if (config.selected_group == devices[dev_num].device_group) {
                        results.push(schedules[ri]);
                    }
                    else if (config.selected_group == "all") {
                        results.push(schedules[ri]);
                    }
                }
                res.json({ schedules: results });
            }
        });
    });
});
app.get('/api/versioncheck', function (req, res) {
    refreshFirmwareList();
    res.sendStatus(204); // success, no content
});
app.delete('/api/schedules/:id', function (req, res) {
    Schedule.remove({ _id: req.params.id }, function (err) {
        if (err) {
            // if the id doesn't exist
            res.sendStatus(404);
        }
        else {
            res.sendStatus(204); // success, no content
        }
    });
});
// API calls for testing
app.get('/api/schedules/run', function (req, res) {
    Schedule.find({}, function (err, results) {
        results.forEach(function (val) {
            val.date = new Date();
            Schedule.update({ _id: val.id }, val, function (err, numAffected) {
            });
        });
        res.sendStatus(204); // success, no content
    });
});
var updateStatus = "";
var updateProgress = 0;
app.get('/api/update', function (req, res) {
    res.send({ status: updateStatus, progress: updateProgress });
});
app.post('/api/update', function (req, res) {
    updateStatus = "Downloading";
    var localFilename = './download.tar.gz';
    var file = fs.createWriteStream(localFilename);
    var request = http.get('http://forum.netonix.com/netonix-manager/netonix-manager-' + req.body.version + ".tar.gz", function (response) {
        if (response.statusCode == 404) {
            file.close();
            fs.unlinkSync(localFilename);
            res.sendStatus(404);
            return;
        }
        res.sendStatus(204); // success, no content
        var str = progress({
            length: response.headers['content-length'],
            time: 100
        });
        str.on('progress', function (progress) {
            updateProgress = progress.percentage;
        });
        response.pipe(str).pipe(file);
        file.on('finish', function () {
            file.close();
            fs.renameSync(localFilename, './update.tar.gz');
            updateStatus = "Updating";
            // the bootloader should now notice that update.tar.gz exists and begin the update, which will kill this process
        });
    }).on('error', function (err) {
        file.close();
        fs.unlinkSync(localFilename);
        res.sendStatus(404);
    });
});
// start refreshing all existing devices
setInterval(function () {
    Schedule.find({}, function (err, results) {
        if (err) {
            util.log("Retrieving Schedule failed, if this continues Schedule data is corrupted and you may want to delete the storage/schedule file and restart the Netonix Manager.");
        }
        else {
            var now_1 = new Date();
            results.forEach(function (schedule) {
                if (schedule.date <= now_1) {
                    util.log("Starting scheduled operation");
                    if (schedule.firmware == "Reboot") {
                        schedule.devices.forEach(function (id) {
                            if (devices[id] != null) {
                                devices[id].reboot();
                            }
                        });
                        util.notify("Scheduled reboot complete, " + results.length + " devices rebooted");
                    }
                    else {
                        var firmwareFile = schedule.firmware.replace("Upgrade to ", "");
                        schedule.devices.forEach(function (id) {
                            if (devices[id] != null) {
                                devices[id].upgrade_firmware_file = firmwareFile;
                            }
                        });
                    }
                    schedule.remove();
                }
            });
        }
    });
    var devices_by_last_query = {};
    var upgrade_count = 0;
    var query_count = 0;
    // we build up a list of devices sorted by the last time it was queried
    // so that the least recently queried devices are at the top
    for (var key in devices) {
        var device = devices[key];
        devices_by_last_query[device.last_query] = parseInt(key);
        if (device.querying) {
            query_count++;
        }
        if (device.upgrading) {
            upgrade_count++;
        }
    }
    for (var key in devices_by_last_query) {
        var device = devices[devices_by_last_query[key]];
        if (device.query()) {
            // we want to limit how many queries are happening at one time
            query_count++;
            if (query_count >= 10) {
                break;
            }
        }
    }
    for (var key in devices) {
        var device = devices[key];
        if (device.upgrade_firmware_file != '' && upgrade_count < 10) {
            device.upgradeFirmware();
            upgrade_count++;
        }
    }
}, 8000);
DeviceInfo.find({}, function (err, deviceinfos) {
    if (err) {
        // this shouldn't happen unless something is wrong with the db
    }
    else {
        deviceinfos.forEach(function (deviceinfo) {
            devices[deviceinfo.id] = new Device(deviceinfo);
        });
    }
});
function refreshFirmwareList() {
    Config.findOne({}, function (err, config) {
        request('http://forum.netonix.com/firmware', function (error, response, body) {
            if (error) {
                util.log("Error getting list of firmwares from the Netonix server");
            }
            else if (response.statusCode == 200) {
                repositoryFirmwares = [];
                var myArray;
                var re = /<a href="(wispswitch-.*bin)">/g;
                while ((myArray = re.exec(body)) !== null) {
                    repositoryFirmwares.push(myArray[1]);
                }
                // sort by version descending
                util.sortVersions(repositoryFirmwares);
                for (var i = 0; i < repositoryFirmwares.length; i++) {
                    if (repositoryFirmwares[i].indexOf("rc") == -1) {
                        var ver = repositoryFirmwares[i].replace("wispswitch-", "").replace(".bin", "");
                        if (ver != config.latest_firmware) {
                            if (config.latest_firmware != '') {
                                util.notify("New firmware version available: " + ver);
                            }
                            config.latest_firmware = ver;
                            config.save();
                        }
                        break;
                    }
                }
            }
        });
        setTimeout(function () {
            request('http://forum.netonix.com/netonix-manager', function (error, response, body) {
                if (error) {
                    util.log("Error getting list of Netonix Managers from the Netonix server");
                }
                else if (response.statusCode == 200) {
                    var nms = [];
                    var myArray;
                    var re = /<a href="netonix-manager-(.*)\.exe">/g;
                    while ((myArray = re.exec(body)) !== null) {
                        nms.push(myArray[1]);
                    }
                    util.sortVersions(nms);
                    if (nms[0] != config.last_seen_version) {
                        if (config.last_seen_version != '' && nms[0] != version) {
                            util.notify("New Netonix Manager version available: " + nms[0]);
                        }
                        config.last_seen_version = nms[0];
                        config.save();
                    }
                }
            });
        }, 500);
    });
}
setInterval(refreshFirmwareList, 3600000); // once an hour
// error handlers need to be last in the handler chain
function errorHandler(err, req, res, next) {
    console.log('error on request %s %s: %j', req.method, req.url, err);
    util.log('Error in ' + req.method + ' ' + req.url + ': ' + err);
    try {
        res.status(500).send("Something bad happened. :(");
    }
    catch (ex) {
    }
}
app.use(errorHandler);
if ('development' == app.get('env')) {
    http.createServer(app).listen(app.get('port'), function () {
        console.log('Express server listening on port ' + app.get('port'));
    });
}
else {
    https.createServer({ key: fs.readFileSync('./private.key'), cert: fs.readFileSync('./certificate.pem') }, app).listen(3443, function () {
        console.log('Express server listening on port 3443');
    });
}
// automatic configuration backup
setInterval(function () {
    for (var key in devices) {
        if (devices[key].uptime > 0 && devices[key].config_version != devices[key].deviceInfo.config_version) {
            devices[key].backupConfig();
            return;
        }
    }
}, 1000);
//# sourceMappingURL=app.js.map