"use strict";
var DeviceInfo = require("./DeviceInfo");
var fs = require("fs");
var path = require("path");
var progress = require('progress-stream');
var util = require("../util");
var ssh2 = require("ssh2");
var ssh2shell = require("ssh2shell");
var request = require('request');
var Config = require("./Config");
var Device = /** @class */ (function () {
    function Device(di) {
        this.monitoring = false;
        this.backing_up = false;
        this.last_query = 0;
        this.upgrade_firmware_file = '';
        this.upgrading = false;
        this.querying = false;
        this.conn_failure_count = 0;
        this.current_failure_count = 0;
        this.dcdc_input_voltage = -1;
        this.id = di._id;
        this.deviceInfo = di;
        this.address = di.address;
        this.ssh_port = di.ssh_port;
        this.web_port = di.web_port;
        this.username = di.username;
        this.password = di.password;
        this.monitor = di.monitor;
        this.name = di.name;
        this.model = di.model;
        this.config_version = di.config_version;
        this.last_seen = di.last_seen;
        this.firmware = di.firmware;
        this.device_group = di.device_group;
        this.uptime = 0;
        this.status = "";
        this.statusProgress = -1;
        this.bootid = '';
        this.bandwidth = 0;
        this.pps = 0;
        this.power = 0;
        this.temp = 0;
        this.web_protocol = '';
        this.https_port = 0;
        this.http_port = 0;
        this.warning = '';
        this.conn_failure_count = 0;
    }
    // returns TRUE if a query is started or was already in progress
    // returns FALSE if not querying and a query was not started or monitoring is disabled
    Device.prototype.query = function () {
        var device = this;
        if (device.querying) {
            return true;
        }
        if ('development' == global["app"].get('env') && this.address.indexOf("192.168.10.") == 0) {
            // this is just for testing
            setTimeout(function () {
                device.model = "WS-12-250A";
                device.firmware = "1.3.9rc13";
                device.name = "Netonix Switch";
                device.status = "";
                device.statusProgress = -1;
                device.uptime = 666;
                device.bootid = "arandombootid";
                device.power = 8.1;
                device.pps = 3122;
                device.bandwidth = 1588;
                device.temp = 55;
                device.config_version = 1;
            }, 1000);
            return true;
        }
        if (!device.monitor) {
            device.status = '';
            device.statusProgress = -1;
            device.warning = '';
            device.bandwidth = 0;
            device.pps = 0;
            device.power = 0;
            device.temp = 0;
            device.config_version = 0;
            device.querying = false;
            device.monitoring = false;
            return false;
        }
        // if we already did a query in the last 10 seconds then don't start another
        if (Date.now() - device.last_query < 10000) {
            return false;
        }
        device.monitoring = true;
        device.querying = true;
        device.last_query = Date.now();
        // its faster to get status updates via HTTP, so do that if possible (web server might be disabled on the switch)
        // but we have to use SSH initially to get the bootid
        if (device.bootid == '' || device.web_protocol == '') {
            var host = {
                disableColorFilter: true,
                server: {
                    host: device.deviceInfo.address,
                    port: device.deviceInfo.ssh_port,
                    userName: device.deviceInfo.username,
                    password: device.deviceInfo.password
                },
                commands: [
                    'cmdline',
                    // we need to ensure that the data we get back from these commands ends with newline so that we can properly remove the command prompt that will follow it
                    // just doing a "cat" won't do that
                    // this grep command is kind of a silly hack that will
                    // Added 9/28/2021 auth Stephen: We are adding 'grep . /switch/config' here because on the WS3 /tmp/config.json does not exist, I would prefer to make this inference from the device.model value, but that does not return until
                    // after we grep /tmp/30sec.json, so we either can make the user specify if this is a WS3 or not, or simply have the manager look for a valid response from attempting to grep /www/config.json and /switch/config.json, then we know which one is correct based on which one responds with valid json data
                    'grep . /tmp/bootid',
                    'grep . /www/config.json',
                    'grep . /switch/config.json',
                    'grep . /tmp/30sec.json',
                    'exit'
                ],
                // if we don't include this null function then it outputs messages to the console
                msg: {
                    send: function (message) {
                    }
                },
                onCommandComplete: function (command, response, sshObj) {
                    response = response.replace(command, "");
                    response = response.replace(/^.*# $/gm, "");
                    response = response.replace('\n', '');
                    if (command === 'grep . /www/config.json') {
                        var result = response.indexOf("No such file or directory");
                        if (result === -1) {
                            try {
                                var config = JSON.parse(response);
                                device.name = config.Switch_Name;
                                device.https_port = parseInt(config.HTTPS_Port);
                                // firmware prior to 1.4.0 don't have this config value
                                if (config.HTTPS_HTTP_Port) {
                                    device.http_port = parseInt(config.HTTPS_HTTP_Port);
                                }
                                else {
                                    device.http_port = 80;
                                }
                                if (config.HTTPS_Allow_HTTP) {
                                    device.web_protocol = "http";
                                }
                                else if (config.HTTPS_Enable) {
                                    device.web_protocol = "https";
                                }
                                else {
                                    device.web_protocol = "";
                                }
                                device.config_version = parseInt(config.Config_Version);
                            }
                            catch (err) {
                                // if we connect to a switch that never had its config saved then config.json will not exist
                                // then JSON.parse() will throw an exception
                                // in that case just use default values
                                device.name = "Netonix Switch";
                                device.https_port = 443;
                                device.http_port = 80;
                                device.web_protocol = "https";
                                device.config_version = 0;
                            }
                        }
                    }
                    else if (command === 'grep . /switch/config.json') {
                        var result = response.indexOf("No such file or directory");
                        if (result === -1) {
                            try {
                                var config = JSON.parse(response);
                                device.name = config.Switch_Name;
                                device.https_port = parseInt(config.HTTPS_Port);
                                // firmware prior to 1.4.0 don't have this config value
                                if (config.HTTPS_HTTP_Port) {
                                    device.http_port = parseInt(config.HTTPS_HTTP_Port);
                                }
                                else {
                                    device.http_port = 80;
                                }
                                if (config.HTTPS_Allow_HTTP) {
                                    device.web_protocol = "http";
                                }
                                else if (config.HTTPS_Enable) {
                                    device.web_protocol = "https";
                                }
                                else {
                                    device.web_protocol = "";
                                }
                                device.config_version = parseInt(config.Config_Version);
                            }
                            catch (err) {
                                // if we connect to a switch that never had its config saved then config.json will not exist
                                // then JSON.parse() will throw an exception
                                // in that case just use default values
                                device.name = "Netonix Switch";
                                device.https_port = 443;
                                device.http_port = 80;
                                device.web_protocol = "https";
                                device.config_version = 0;
                            }
                        }
                    }
                    else if (command === 'grep . /tmp/bootid') {
                        device.bootid = response;
                    }
                    else if (command === 'grep . /tmp/30sec.json') {
                        if (device.monitor || device.model == '' || device.firmware == '') {
                            updateDeviceStats(device, response);
                        }
                    }
                },
                onEnd: function (sessionText, sshObj) {
                    device.querying = false;
                    device.conn_failure_count = 0;
                    DeviceInfo.update({ _id: device.id }, { name: device.name, model: device.model, firmware: device.firmware }, function (err, numAffected) {
                    });
                },
                onError: function (err, type, close, callback) {
                    if (device.model.indexOf("WS3") !== -1 && device.upgrading) {
                        device.rebooting();
                    }
                    else {
                        device.querying = false;
                        if (err.message === 'All configured authentication methods failed') {
                            device.status = "Invalid username or password";
                        }
                        else {
                            device.connectionError("Unable to open SSH connection");
                        }
                        if (callback) {
                            callback(err, type);
                        }
                        if (close) {
                            return this.connection.end();
                        }
                    }
                }
            };
            new ssh2shell(host).connect();
        }
        else {
            var port = (device.web_protocol == 'https' ? device.https_port : device.http_port);
            if (device.web_port != '') {
                port = device.web_port;
            }
            var reqUrl = device.web_protocol + "://" + device.address + ":" + port + "/api/v1/status/30sec?" + device.bootid;
            var req = request({ url: reqUrl, agentOptions: { rejectUnauthorized: false }, timeout: 5000 }, function (error, response, body) {
                if (error) {
                    if (device.model.indexOf("WS3") !== -1 && device.upgrading) {
                        device.rebooting();
                    }
                    else {
                        device.connectionError(error);
                    }
                }
                else if (response.statusCode == 200) {
                    updateDeviceStats(device, body);
                    device.conn_failure_count = 0;
                }
                device.querying = false;
            });
        }
        return true;
    };
    Device.prototype.rebooting = function () {
        var _this = this;
        this.status = "Rebooting";
        this.statusProgress = -1;
        this.querying = false;
        this.upgrading = false;
        this.backing_up = false;
        this.uptime = 0;
        if (!this.monitor) {
            setTimeout(function () {
                _this.status = '';
            }, 10000);
        }
    };
    Device.prototype.upgradeFirmware = function () {
        var device = this;
        var firmwareFile = device.upgrade_firmware_file;
        device.upgrade_firmware_file = '';
        device.upgrading = true;
        util.log("Upgrading device " + device.address + " with " + firmwareFile);
        if ('development' == global["app"].get('env') && this.address.indexOf("192.168.10.") == 0) {
            var n = 0;
            var timer = setInterval(function () {
                if (n == 0) {
                    device.status = "Uploading firmware";
                    device.statusProgress = 0;
                }
                else if (n == 1) {
                    device.statusProgress = 25;
                }
                else if (n == 2) {
                    device.statusProgress = 50;
                }
                else if (n == 3) {
                    device.statusProgress = 75;
                }
                else if (n == 4) {
                    device.status = "Upgrading firmware";
                    device.statusProgress = 0;
                }
                else if (n == 5) {
                    device.statusProgress = 25;
                }
                else if (n == 6) {
                    device.statusProgress = 50;
                }
                else if (n == 7) {
                    device.statusProgress = 75;
                }
                else if (n == 8) {
                    util.log("Upgrading device " + device.address + " complete, rebooting");
                    device.rebooting();
                    device.upgrading = false;
                }
                else if (n == 9) {
                    device.status = "";
                    device.statusProgress = -1;
                    device.uptime = 30;
                    clearInterval(timer);
                }
                n++;
            }, 500);
            return;
        }
        var filename = global["datadir"] + '/firmwares/' + firmwareFile;
        if (!fs.existsSync(filename)) {
            util.notify("firmware file " + filename + " is missing", device);
            device.upgrading = false;
            return;
        }
        // WS upgrade commands
        var uploadCmd = 'scp -t /tmp/upgrade.bin';
        // calling firmware_upgrade like this makes it run in the background, so it will finish even if this session disconnects
        var upgradeCmd = 'cmdline\n ((firmware_upgrade /tmp/upgrade.bin)&)&\n PID=`pgrep firmware_upgrad -o`\n kill -0 $PID 2>/dev/null\n while [ $? == 0 ]; do sleep 1; kill -0 $PID 2>/dev/null; done\nkillall ssmtp;exit\nexit\n';
        if (filename.indexOf("ws3") !== -1) {
            // this is a WS3 update, update commands accordingly
            uploadCmd = 'scp -t /tmp/upgrade.mfi';
        }
        var conn = new ssh2();
        conn.on('ready', function () {
            conn.exec(uploadCmd, function (err, stream) {
                if (err) {
                    util.notify("error copying firmware to remote", device);
                    device.upgrading = false;
                    return;
                }
                device.status = "Uploading firmware";
                device.statusProgress = 0;
                var stat = fs.statSync(filename);
                stream.write("C0644 " + stat.size + " " + firmwareFile + "\n");
                var str = progress({
                    length: stat.size,
                    time: 100
                });
                str.on('progress', function (progress) {
                    device.statusProgress = progress.percentage;
                });
                var rs = fs.createReadStream(filename);
                rs.pipe(str).pipe(stream);
                stream.on('finish', function (data) {
                    // this will be used for WS monitoring
                    var buffer = '';
                    var errorOccured = false;
                    function updateStatus(data) {
                        var text = (buffer + data.toString()).replace("\r\n", "\n");
                        var a = text.indexOf('\n');
                        while (a > -1) {
                            var msg = text.substr(0, a);
                            if (msg.indexOf("Error") > -1) {
                                util.notify("error during upgrade: " + msg.replace("Error:", ""), device);
                                errorOccured = true;
                            }
                            else if (msg.indexOf("...") > -1 || msg.indexOf("%") > -1) {
                                if (msg.indexOf("%") > 0) {
                                    device.statusProgress = parseInt(msg);
                                }
                                else {
                                    device.status = msg;
                                    device.statusProgress = -1;
                                }
                            }
                            text = text.substr(a + 1);
                            a = text.indexOf('\n');
                        }
                        buffer = text;
                        if (text.indexOf("...") > -1 || text.indexOf("%") > -1) {
                            if (text.indexOf("%") > 0) {
                                device.statusProgress = parseInt(text);
                            }
                            else {
                                device.status = text;
                                device.statusProgress = -1;
                            }
                        }
                    }
                    if (firmwareFile.indexOf("ws3") === -1) {
                        try {
                            // if the connection closed during the above command will we still get here but conn won't be valid any more
                            // in that case conn.shell will throw an exception
                            conn.shell(function (err, stream) {
                                if (err) {
                                    util.notify("error starting remote shell", device);
                                    device.upgrading = false;
                                    return;
                                }
                                device.status = "Upgrading firmware";
                                device.statusProgress = 0;
                                stream.on('finish', function () {
                                    conn.end();
                                    device.statusProgress = -1;
                                    if (errorOccured) {
                                        device.status = "";
                                    }
                                    else {
                                        util.log("Upgrading device " + device.address + " complete, rebooting");
                                        device.rebooting();
                                    }
                                    device.upgrading = false;
                                }).on('data', function (data) {
                                    updateStatus(data);
                                });
                                stream.write(upgradeCmd);
                            });
                        }
                        catch (err) {
                            util.notify("connection unexpectedly closed during upgrade:" + err, device);
                            device.upgrading = false;
                        }
                    }
                    else {
                        // ssh2 simply doesn't work for WS3, but ssh2shell seems to work fine, so let's just build out the commands to make that work
                        var host = {
                            disableColorFilter: true,
                            server: {
                                host: device.deviceInfo.address,
                                port: device.deviceInfo.ssh_port,
                                userName: device.deviceInfo.username,
                                password: device.deviceInfo.password
                            },
                            commands: [
                                'cmdline',
                                '(netonix_app -u) &'
                            ],
                            // if we don't include this null function then it outputs messages to the console
                            onCommandComplete: function (command, response, sshObj) {
                                if (command === '(netonix_app -u) &') {
                                    console.log(response);
                                }
                            },
                            onEnd: function (sessionText, sshObj) {
                                util.log("Upgrading device " + device.address + " in progress, rebooting soon");
                                device.status = "Upgrading Firmware";
                                device.statusProgress = -1;
                            },
                            onError: function (err, type, close, callback) {
                                console.log(err.message);
                                if (callback) {
                                    callback(err, type);
                                }
                                if (close) {
                                    return this.connection.end();
                                }
                            }
                        };
                        new ssh2shell(host).connect();
                    }
                });
            });
        }).on('error', function (err) {
            util.notify("error connecting to switch:" + err, device);
            device.upgrading = false;
        }).connect({
            host: device.deviceInfo.address,
            port: device.deviceInfo.ssh_port,
            username: device.deviceInfo.username,
            password: device.deviceInfo.password,
            readyTimeout: 10000
        });
    };
    Device.prototype.reboot = function () {
        var device = this;
        util.log("Rebooting device " + device.address);
        var host = {
            disableColorFilter: true,
            server: {
                host: device.deviceInfo.address,
                port: device.deviceInfo.ssh_port,
                userName: device.deviceInfo.username,
                password: device.deviceInfo.password
            },
            commands: [
                'cmdline',
                'killall ssmtp',
                'reboot',
                'exit'
            ],
            // if we don't include this null function then it outputs messages to the console
            msg: {
                send: function (message) {
                }
            },
            onEnd: function (sessionText, sshObj) {
                device.rebooting();
            },
            onError: function (err, type, close, callback) {
                util.log(device.address + ": Error rebooting");
                if (callback) {
                    callback(err, type);
                }
                if (close) {
                    return this.connection.end();
                }
            }
        };
        new ssh2shell(host).connect();
    };
    Device.prototype.connectionError = function (msg) {
        var _this = this;
        Config.findOne({}, function (err, config) {
            var device = _this;
            var counts_till_offline = 2;
            if (err) {
                counts_till_offline = 0;
            }
            else {
                counts_till_offline = config.count_until_offline;
            }
            if (device.uptime > 0 && device.conn_failure_count >= counts_till_offline) {
                util.notify("changed state to Offline: " + msg, device);
                device.statusProgress = -1;
                device.querying = false;
                device.warning = '';
                device.bootid = '';
                device.bandwidth = 0;
                device.pps = 0;
                device.power = 0;
                device.temp = 0;
                device.web_protocol = '';
                device.https_port = 0;
                device.http_port = 0;
                device.config_version = 0;
                device.last_seen = Date.now();
                device.status = "";
                device.uptime = 0;
                DeviceInfo.update({ _id: device.id }, { last_seen: device.last_seen }, function (err, numAffected) {
                });
            }
            else {
                device.conn_failure_count += 1;
            }
            if (device.status != "Rebooting") {
                device.status = '';
            }
        });
    };
    Device.prototype.backupConfig = function () {
        if (this.backing_up) {
            return;
        }
        var device = this;
        if ('development' == global["app"].get('env') && this.address.indexOf("192.168.10.") == 0) {
            // this is just for testing
            DeviceInfo.update({ _id: device.id }, { config_version: device.config_version }, function (err, numAffected) {
            });
            return;
        }
        this.backing_up = true;
        var data = "";
        util.log("Backing up configuration for device " + device.address);
        var host = {
            disableColorFilter: true,
            server: {
                host: device.deviceInfo.address,
                port: device.deviceInfo.ssh_port,
                userName: device.deviceInfo.username,
                password: device.deviceInfo.password
            },
            commands: [
                'cmdline',
                'config_backup',
                'openssl enc -base64 -in /tmp/wispswitch_config.tar',
                'exit'
            ],
            // if we don't include this null function then it outputs messages to the console
            msg: {
                send: function (message) {
                }
            },
            onData: function (newdata) {
                data += newdata;
            },
            onCommandComplete: function (command, response, sshObj) {
                if (command === 'openssl enc -base64 -in /tmp/wispswitch_config.tar') {
                    data = data.substr(data.indexOf(command) + command.length + 2);
                    data = data.replace(/^.*# $/gm, "");
                    data = data.replace('\n', '');
                    data = data.replace('\r', '');
                    var filename = device.address + "_" + device.model + "_" + device.name.replace(/[^a-z0-9\._-]+/gi, "-") + "_" + device.config_version + ".ncfg";
                    var buf = new Buffer(data, 'base64');
                    fs.writeFileSync(path.join(global["datadir"], 'configs/' + filename), buf);
                    DeviceInfo.update({ _id: device.id }, { config_version: device.config_version }, function (err, numAffected) {
                        device.deviceInfo.config_version = device.config_version;
                        device.backing_up = false;
                        util.log(device.address + ": Backed up configuration to " + filename);
                        // purge old configs
                        var files = fs.readdirSync(path.join(global["datadir"], 'configs'));
                        var myconfigs = {};
                        for (var i = 0; i < files.length; i++) {
                            if (files[i].indexOf(device.address + "_") == 0) {
                                var m = files[i].match(/_(\d+).ncfg/);
                                if (m != null) {
                                    myconfigs[m[1]] = files[i];
                                }
                            }
                        }
                        while (Object.keys(myconfigs).length > 5) {
                            var key = Object.keys(myconfigs)[0];
                            fs.unlinkSync(path.join(global["datadir"], 'configs/' + myconfigs[key]));
                            delete myconfigs[key];
                        }
                    });
                }
            },
            onError: function (err, type, close, callback) {
                util.log(device.address + ": Error backing up config");
                device.backing_up = false;
                if (callback) {
                    callback(err, type);
                }
                if (close) {
                    return this.connection.end();
                }
            }
        };
        new ssh2shell(host).connect();
    };
    Device.prototype.getLogAlerts = function (resultCallback) {
        var device = this;
        var host = {
            disableColorFilter: true,
            server: {
                host: device.deviceInfo.address,
                port: device.deviceInfo.ssh_port,
                userName: device.deviceInfo.username,
                password: device.deviceInfo.password
            },
            commands: [
                'cmdline',
                'rm /tmp/unseenlog',
                'cat /var/log/messages; if [ -e /var/log/messages.0 ]; then cat /var/log/messages.0; fi',
                'exit'
            ],
            // if we don't include this null function then it outputs messages to the console
            msg: {
                send: function (message) {
                }
            },
            onCommandComplete: function (command, response, sshObj) {
                response = response.replace(command, "");
                response = response.replace(/^.*# $/gm, "");
                response = response.replace('\n', '');
                if (command === 'cat /var/log/messages; if [ -e /var/log/messages.0 ]; then cat /var/log/messages.0; fi') {
                    var lines = response.split(/\n/).filter(function (item) { return item.indexOf(": !") > 0; });
                    resultCallback(lines.join("\n"));
                }
            },
            onError: function (err, type, close, callback) {
                resultCallback("");
                if (callback) {
                    callback(err, type);
                }
                if (close) {
                    return this.connection.end();
                }
            }
        };
        new ssh2shell(host).connect();
    };
    return Device;
}());
;
function updateDeviceStats(device, statsjson) {
    try {
        var stats = JSON.parse(statsjson);
        if (stats.Uptime == 0) {
            // if we get here it means that 30sec is only returning Uptime:0 because the bootid is wrong so do a full update
            device.bootid = '';
            return;
        }
        device.model = stats.Model;
        device.firmware = stats.Firmware_Version;
        if (device.monitor) {
            if (device.uptime == 0 && stats.Uptime > 0) {
                util.notify("changed state to Online", device);
                device.status = '';
            }
            device.uptime = stats.Uptime;
            device.temp = stats.Board_Temp;
            device.bandwidth = 0;
            device.bandwidth += stats.UsageData[0].txThroughputPerPort.reduce(function (a, b) { return a + b; });
            device.bandwidth += stats.UsageData[0].rxThroughputPerPort.reduce(function (a, b) { return a + b; });
            device.pps = 0;
            device.pps += stats.UsageData[0].txPacketRatePerPort.reduce(function (a, b) { return a + b; });
            device.pps += stats.UsageData[0].rxPacketRatePerPort.reduce(function (a, b) { return a + b; });
            device.power = stats.BasePowerUsage;
            device.power += stats.UsageData[0].powerPerPort.reduce(function (a, b) { return a + b; });
            if (stats.hasOwnProperty("DCDC_Input_Voltage")) {
                device.dcdc_input_voltage = stats.DCDC_Input_Voltage;
            }
            else {
                device.dcdc_input_voltage = -1;
            }
            if (stats.UnseenLogMessages) {
                device.warning = "Log alerts";
            }
            else {
                device.warning = '';
            }
            if (device.config_version != stats.Config_Version) {
                // the config changed since we last saw it
                // need to do a full update
                device.bootid = '';
            }
            device.config_version = stats.Config_Version;
        }
    }
    catch (err) {
        console.log("updateDeviceStats:" + device.address);
        console.log(statsjson);
    }
}
module.exports = Device;
//# sourceMappingURL=Device.js.map